// Project Sword & Gun Copyright © 2021 Kevin Poretti


#include "Components/SNGCharacterMovementComponent.h"

#include "Kismet/KismetMathLibrary.h"
#include "PhysicsEngine/PhysicsSettings.h"

USNGCharacterMovementComponent::USNGCharacterMovementComponent()
{
	// Defaults
	DefaultGroundFriction = 8.0f;
	DefaultMaxWalkSpeed = 600.0f;
	DefaultRotationRate = FRotator(0.0f, 360.0f, 0.0f);
	
	// sprint
	SprintingMaxWalkSpeed = 1000.0f;
	SprintLateralSpeedThreshold = 500.0f;
	SprintingRotationRate = FRotator(0.0f, 180.0f, 0.0f);
	TimeToInitiateSprint = 2.0f;
	WalkToSprintInterpSpeed = 4.0f;

	// firing
	FiringMaxWalkSpeed = 300.0f;
	
	// reloading
	ReloadingMaxWalkSpeed = 300.0f;
	
	// airborne defaults
	DefaultAirControl = 0.3f;
	
	// jump
	MaxJumpHeight = 250.0f;
	DesiredJumpDuration = 0.75f;
	ComputerGravityScaleAndJumpZVelocity();

	/*
	// not used...yet
	MinJumpHeight = 200.0f;
	MaxJumpTime = 0.2f;

	LateralVelocityDampening = 0.8f;
	
	MinAllowedDashHeight = 175.0f;
	MinAllowedAttackHeight = 175.0f;
	*/

	// Character Movement component defaults
	SetWalkableFloorAngle(45.622231f);
	
	AirControlBoostMultiplier = 0.0f;
	AirControlBoostVelocityThreshold = 0.0f;
}

/**
 *	Instantly sets all movement parameters to their defaults values (i.e. ground friction to DefaultGroundFriction,
 *	MaxWalkSpeed to DefaultMaxWalkSpeed, etc.)
 */
void USNGCharacterMovementComponent::RestoreMovementDefaults()
{
	GroundFriction = DefaultGroundFriction;
	GravityScale = DefaultGravityScale;
	RotationRate = DefaultRotationRate;
	MaxWalkSpeed = DefaultMaxWalkSpeed;
	AirControl = DefaultAirControl;
	JumpZVelocity = DefaultJumpZVelocity;
}

#if WITH_EDITOR
void USNGCharacterMovementComponent::PostEditChangeProperty(FPropertyChangedEvent& PropertyChangedEvent)
{
	Super::PostEditChangeProperty(PropertyChangedEvent);

	const FProperty* PropertyThatChanged = PropertyChangedEvent.MemberProperty;
	if (PropertyThatChanged && PropertyThatChanged->GetFName() == GET_MEMBER_NAME_CHECKED(USNGCharacterMovementComponent, MaxJumpHeight) ||
		PropertyThatChanged && PropertyThatChanged->GetFName() == GET_MEMBER_NAME_CHECKED(USNGCharacterMovementComponent, DesiredJumpDuration))
	{
		// Compute WalkableFloorZ from the Angle.
		ComputerGravityScaleAndJumpZVelocity();
	}
}

void USNGCharacterMovementComponent::PostLoad()
{
	Super::PostLoad();

	// Compute gravity scale and jump z velocity for first time
	ComputerGravityScaleAndJumpZVelocity();
}
#endif // WITH_EDITOR

void USNGCharacterMovementComponent::OnSprintTimer()
{
	bIsSprinting = true;
}

void USNGCharacterMovementComponent::ComputerGravityScaleAndJumpZVelocity()
{
	float HalfJumpDuration = DesiredJumpDuration / 2.0f;
	DefaultJumpZVelocity = (2 * MaxJumpHeight) / HalfJumpDuration;
	
	float GameGravity = UPhysicsSettings::Get()->DefaultGravityZ;
	float TargetGravity = (-2 * MaxJumpHeight) / (HalfJumpDuration * HalfJumpDuration);
	DefaultGravityScale = TargetGravity / GameGravity;
}

void USNGCharacterMovementComponent::StartJump()
{
	/*
	JumpTimer = MaxJumpTime;

	Velocity = FVector(Velocity.X * LateralVelocityDampening, Velocity.Y * LateralVelocityDampening, 0.0f);
	*/
}

void USNGCharacterMovementComponent::AddJumpForce()
{
	/*
	if(JumpTimer > 0.0f)
	{
		Velocity = FVector(Velocity.X, Velocity.Y, Velocity.Z + (JumpForce * GetWorld()->DeltaTimeSeconds));

		JumpTimer = FMath::Max(JumpTimer - GetWorld()->DeltaTimeSeconds, 0.0f);
	}
	*/
}

void USNGCharacterMovementComponent::BeginPlay()
{
	Super::BeginPlay();

	// set character movement parameters to defaults
	RestoreMovementDefaults();
}

void USNGCharacterMovementComponent::TickComponent(float DeltaTime, ELevelTick Tick,
	FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, Tick, ThisTickFunction);

	float LateralSpeed = Velocity.Size2D();
	if(LateralSpeed >= SprintLateralSpeedThreshold && !GetWorld()->GetTimerManager().IsTimerActive(TimerHandle_SprintTimer) )
	{
		GetWorld()->GetTimerManager().SetTimer(TimerHandle_SprintTimer,
			this, &USNGCharacterMovementComponent::OnSprintTimer,
			TimeToInitiateSprint, false);
	}

	if(LateralSpeed < SprintLateralSpeedThreshold && GetWorld()->GetTimerManager().IsTimerActive(TimerHandle_SprintTimer))
	{
		GetWorld()->GetTimerManager().ClearTimer(TimerHandle_SprintTimer);
		bIsSprinting = false;
	}
	
	// set walk speed and rotation rate depending on whether player is sprinting or not
	float TargetMaxWalkSpeed = bIsSprinting ? SprintingMaxWalkSpeed : DefaultMaxWalkSpeed;
	float NewMaxWalkSpeed = FMath::FInterpTo(MaxWalkSpeed, TargetMaxWalkSpeed, DeltaTime, WalkToSprintInterpSpeed);
	MaxWalkSpeed = NewMaxWalkSpeed;

	// NOTE(kevin): since default rotate rate is 360 RInterpTo will go from 0 to 180 instead of 360 to 180
	// need to think of a way to prevent this
	// might just need to do my own interp manually
	/*
	FRotator TargetRotationRate = bIsSprinting ? SprintingRotationRate : DefaultRotationRate;
	FRotator NewRotationRate = FMath::RInterpTo(RotationRate, TargetRotationRate, DeltaTime, WalkToSprintInterpSpeed);
	RotationRate = NewRotationRate;
	*/
}
